package va.montecarlo;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;

import org.apache.log4j.Logger;

import va.calendar.SDate;
import va.curve.*;
import va.policy.Policy;
import va.scenario.FundMap;
import va.scenario.Scenario;

public abstract class Pricer {
	protected String name;
	protected Logger log = Logger.getLogger(Pricer.class);
	protected Scenario scenario;
	protected int scenarioOffset;
	
	protected int numScenario;
	protected int numTimeStep;
	protected double dT;
	
	protected MortalityCurve mcfemale;
	protected MortalityCurve mcmale;
	protected double[] fc;
	
	//cash flow matrix
	protected double[][] AV;  // account value after all events
	protected double[][] DA; // payoff of death benefit
	protected double[][] LA; // payoff of living benefit
	protected double[][] RC;  // risk charge
	
	protected double[] df;
	protected double[] q;
	protected double[] s;
	
	
	public Pricer(Scenario scenario, int scenarioOffset) {
		this.scenario = scenario;
		this.scenarioOffset = scenarioOffset;

		numScenario = scenario.getNumScenario();
		numTimeStep = scenario.getNumTimeStep();
		AV = new double[numScenario][numTimeStep-scenarioOffset];
		DA = new double[numScenario][numTimeStep-scenarioOffset];
		LA = new double[numScenario][numTimeStep-scenarioOffset];
		RC = new double[numScenario][numTimeStep-scenarioOffset];
		df = new double[numTimeStep+1-scenarioOffset];
		q = new double[numTimeStep+1-scenarioOffset];
		s = new double[numTimeStep+1-scenarioOffset];
		dT  = 1/12.0;
	}
	
	public double[][] getDA() {
		return DA;
	}
	
	public double[][] getLA() {
		return LA;
	}
	
	public double[][] getRC() {
		return RC;
	}
	
	public double[] getDF() {
		return df;
	}
	
	public void setValue(double[] fc, MortalityCurve mcfemale, 
			MortalityCurve mcmale) {
		this.mcfemale = mcfemale;
		this.mcmale = mcmale;
		this.fc = fc;
	}
	
	public Result price(Policy po, Map<Integer, Double> shock, double[] fc, MortalityCurve mcfemale, 
			MortalityCurve mcmale, FundMap fundMap) {
		this.mcfemale = mcfemale;
		this.mcmale = mcmale;
		this.fc = fc;
		
		Policy p = po.clone();
		Result res = new Result();
				
		int numIndex = fundMap.getNumIndex();
		double[] av = new double[numIndex + 1];
		av[numIndex] = 0.0;
		for(int i=0; i<numIndex; ++i) {
			double dSum = 0.0;
			for(int k=0; k<p.vFundNum.length; ++k) {
				dSum += p.vFundValue[k] * ( 1 + shock.get(p.vFundNum[k])) * fundMap.get(p.vFundNum[k])[i];
			}
			av[i] = dSum;
			av[numIndex] += dSum;
		}
		res.av = av;
		
		//int numTimeStep = scenario.getNumTimeStep();
		int T = Math.min( SDate.monthsBetween(p.currentDate, p.matDate), numTimeStep);		
		
		if(T<0) {
			return res;
		}
				
		df[0] = 1.0;
		q[0] = 0.0;
		s[0] = 1.0;
		
		double dAge = SDate.yearsBetween(p.birthDate, p.currentDate);
		for(int j=1; j<numTimeStep+1-scenarioOffset; ++j) {					
			df[j] = df[j-1] * Math.exp(-fc[j-1+scenarioOffset] * dT);			
		}
		if(p.gender.toLowerCase().equals("f")) {
			for(int j=1; j<numTimeStep+1-scenarioOffset; ++j) {					
				q[j] = mcfemale.q(dAge + (j-1) * dT, dT);
				s[j] = mcfemale.p(dAge, j * dT);
			}
		} else {
			for(int j=1; j<numTimeStep+1-scenarioOffset; ++j) {				
				q[j] = mcmale.q(dAge + (j-1) * dT, dT);
				s[j] = mcmale.p(dAge, j * dT);						
			}
			
		}
		
		for(int i=0; i<numScenario; ++i) {
			Policy pc = p.clone();
			for(int k=0; k<pc.vFundNum.length; ++k) {
				pc.vFundValue[k] *= 1 + shock.get(pc.vFundNum[k]);
			}
			for(int j=0; j<numTimeStep - scenarioOffset; ++j) {
				project(pc, i, j);
			}
		}
				
		double pvRC = 0.0;
		double pvDA = 0.0;
		double pvLA = 0.0;
		for(int j=1; j<numTimeStep+1-scenarioOffset; ++j) {
			double dRC = 0.0;
			double dDA = 0.0;
			double dLA = 0.0;
			for(int i=0; i<numScenario; ++i) {
				dRC += RC[i][j-1];
				dDA += DA[i][j-1];
				dLA += LA[i][j-1];
			}
			pvRC += s[j] * dRC * df[j] / numScenario;
			pvDA += s[j-1] * dDA * q[j] * df[j] / numScenario;
			pvLA += s[j] * dLA * df[j] / numScenario;
		}
				
		res.riskCharge = pvRC * p.survivorShip;
		res.gmdb = pvDA * p.survivorShip;
		res.gmlb = pvLA * p.survivorShip;
		res.guaranteePay = (pvDA + pvLA) * p.survivorShip;
		res.fmv = (pvDA + pvLA - pvRC) * p.survivorShip;
		return res;
	}
	
	public abstract void project(Policy p, int sceInd, int timeInd);
	
	double getAnnuityFactor(Policy p, int timeInd) {
		double aT = 0.0;
		double dAge = SDate.yearsBetween(p.birthDate, p.currentDate);
		double dP = 1.0;
		int nY = 0;
		while(dP > 1e-4) {
			double dFR = 0.0;
			if(timeInd + scenarioOffset + nY * 12 < numTimeStep) {
				dFR = fc[timeInd + scenarioOffset + nY * 12];
			} else {
				dFR = fc[numTimeStep-1];
			}
			if(p.gender.toLowerCase().equals("f")) {
				dP = mcfemale.p(dAge, nY);
			} else {
				dP = mcmale.p(dAge, nY);
			}
			aT += dP * Math.exp(-dFR * nY);
			nY += 1;			
		}		
		
		return aT;
	}
	
	double getAnnuityFactor2(Policy p, double r) {
		double aT = 0.0;
		double dAge = SDate.yearsBetween(p.birthDate, p.currentDate);
		double dP = 1.0;
		int nY = 0;
		while(dP > 1e-4) {			
			if(p.gender.toLowerCase().equals("f")) {
				dP = mcfemale.p(dAge, nY);
			} else {
				dP = mcmale.p(dAge, nY);
			}
			aT += dP * Math.exp(-r * nY);
			nY += 1;			
		}		
		
		return aT;
	}

	protected void save(String filename, double[][] M) throws IOException {
		String separator = System.getProperty("line.separator");
		StringBuilder sb = new StringBuilder();
		for(int i=0; i<M.length; ++i) {
			sb.append(M[i][0]);
			for(int k=1; k<M[0].length; ++k) {
				sb.append(',').append(M[i][k]);
			}
			sb.append(separator);
		}
		
		FileWriter outFile = new FileWriter(filename);
		PrintWriter out = new PrintWriter(outFile);
		out.print(sb.toString());
		out.close();

	}
	
	public void saveAV(String postfix) throws IOException {
		save("AV" + postfix + ".csv", AV);
	}
	
	public void saveDA(String postfix) throws IOException {
		save("DA" + postfix + ".csv", DA);
	}
	
	public void saveLA(String postfix) throws IOException {
		save("LA" + postfix + ".csv", LA);
	}
	
	public void saveRC(String postfix) throws IOException {
		save("RC" + postfix + ".csv", RC);
	}
}
