package va.policy;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;

import javax.xml.parsers.*;

import org.apache.commons.math3.random.MersenneTwister;
import org.apache.log4j.Logger;
import org.w3c.dom.*;

import va.calendar.*;
import va.scenario.*;
import va.util.DoubleIntPair;

public class PolicyGenerator {
	protected Logger log = Logger.getLogger(PolicyGenerator.class.getName());
	protected Map<String, String> mapParameters;
	protected SDate birthStartDate;
	protected SDate birthEndDate;
	protected SDate issueStartDate;
	protected SDate issueEndDate;
	protected int minMaturity;
	protected int maxMaturity;
	protected double minAccountValue;
	protected double maxAccountValue;	
	protected double femalePercent;
	protected String outputFile;
	
	protected Map<String, Map<String, Double>> mapProductTypes;	
	protected List<Integer> listFundNumber;
	protected List<Double> listFundFee;
	
	protected String paramFile;
	protected int seed;
	protected MersenneTwister mt;
	protected Inforce inforce;
	protected PolicyAger pa;
	
	public PolicyGenerator(String paramFile) {
		this.paramFile = paramFile;	
	}
	
	public void generate() throws Exception {
		loadParameter();
		FundMap fundMap = new FundMap(getParam("fundmapfile"));
		Scenario scenario = new Scenario(getParam("scenariofolder"), fundMap,
				1, getIntParam("numtimestep"));
		pa = new PolicyAger(scenario, issueStartDate);
		
		inforce = new Inforce();
		SDate currentDate = getDateParam("currentDate");
		log.info(currentDate.toString());
		
		int id = 0;
		for(String prodType : mapProductTypes.keySet()) {
			mt = new MersenneTwister(seed + prodType.hashCode());
			int numPolicy =  (int)getProductTypeParam(prodType,"numpolicy");
			for(int i=0; i<numPolicy; ++i) {
				Policy p = generateOne(++id, prodType);	
				pa.agePolicy(p, currentDate);
				inforce.addPolicy(p);
			}
		}
		
		inforce.saveToCSV(outputFile);
		
	}
	
	public Policy generateOne(int recordID, String prodType) throws Exception {
		Policy p = new Policy();
		p.recordID = recordID;		
		p.survivorShip = 1.0;
		
		// generate gender
		double du = mt.nextDouble();
		if(du<femalePercent) {
			p.gender = "F";
		} else {
			p.gender = "M";
		}
		p.productType = prodType.toUpperCase();
		p.baseFee = getDoubleParam("basefee");	
		p.riderFee = getProductTypeParam(prodType,"riderfee");

		// generate account value
		du = mt.nextDouble();
		double av = (1-du) * minAccountValue + du * (maxAccountValue);
		
		p.gbAmt = av;
		if(prodType.contains("wb")) {
			p.gmwbBalance = av;
			p.wbWithdrawalRate = getProductTypeParam(prodType,"withdrawalrate");
		}
		if(prodType.contains("ru")) {
			p.rollUpRate = getProductTypeParam(prodType,"rolluprate");
		}
			
		// generate birth date		
		int months = SDate.monthsBetween(birthStartDate, birthEndDate);
		int m = mt.nextInt(months);
		p.birthDate = new SDate(birthStartDate);
		p.birthDate.addMonths(m);
			
		// generate issue date
		months =  SDate.monthsBetween(issueStartDate,issueEndDate);
		m = mt.nextInt(months);
		p.issueDate = new SDate(issueStartDate);
		p.issueDate.addMonths(m);
		
		p.currentDate = new SDate(p.issueDate);
		
		// generate maturity date
		int years = maxMaturity - minMaturity;
		int y;
		if(years == 0) {
			y = minMaturity;
		} else {
			y = mt.nextInt(years) + minMaturity;
		}
		SDate date = new SDate(p.issueDate);
		date.setYear(p.issueDate.year() + y);
		p.matDate = date;		
		
		// generate funds
		List<DoubleIntPair> listPair = new ArrayList<DoubleIntPair>();		
		for (int i = 0; i < listFundNumber.size(); i++) {
			listPair.add(new DoubleIntPair(mt.nextDouble(), i));
		}
		Collections.sort(listPair);
			
		p.vFundNum = new int[listFundNumber.size()];
		p.vFundFee = new double[listFundNumber.size()];
		p.vFundValue = new double[listFundNumber.size()];
		for(int i=0; i<listFundNumber.size(); ++i) {
			p.vFundValue[i] = 0.0;
			p.vFundNum[i] = listFundNumber.get(i);
			p.vFundFee[i] = listFundFee.get(i);
		}
		int count = mt.nextInt(listFundNumber.size()) + 1;
		
		int k;
		for(int i=0; i<count;++i) {
			k = listPair.get(i).getIndex();				
			p.vFundValue[k] = av / count;			
		}

		for(int i=0; i<p.vFundNum.length; ++i) {
			if(p.vFundValue[i] <0) {
				throw new Exception(String.format("negative fund value %d, %f", p.vFundNum[i], p.vFundValue[i]));	
			}
		}
		return p;
	}
	
	protected void loadParameter() throws Exception {
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
        Document doc = docBuilder.parse (new File(paramFile));
        
        doc.getDocumentElement().normalize();
        NodeList nl = doc.getElementsByTagName("Parameters");
        if(nl.getLength() != 1) {
        	throw new Exception("Number of Parameters tag != 1");
        }
        
        mapParameters = new HashMap<String, String>();
        NodeList nl2 = ((Element)nl.item(0)).getElementsByTagName("Parameter");
        for(int i=0; i<nl2.getLength(); ++i) {
        	Node node = nl2.item(i);
        	String name = node.getAttributes().getNamedItem("name").getNodeValue();
        	String value = node.getAttributes().getNamedItem("value").getNodeValue();
        	
        	mapParameters.put(name.toLowerCase(), value);
        }
  
        birthStartDate = getDateParam("birthStartDate");
        birthEndDate = getDateParam("birthEndDate");
        issueStartDate = getDateParam("issueStartDate");
        issueEndDate = getDateParam("issueEndDate");
        minMaturity = getIntParam("minMaturity");
        maxMaturity = getIntParam("maxMaturity");
        minAccountValue = getDoubleParam("minAccountValue");
        maxAccountValue = getDoubleParam("maxAccountValue");        
        seed = getIntParam("seed");
        femalePercent = getDoubleParam("femalePercent");
        outputFile = getParam("outputFile");
        
        mapProductTypes = new HashMap<String, Map<String, Double>>();
        nl = doc.getElementsByTagName("ProductTypes");
        if(nl.getLength() != 1) {
        	throw new Exception("Number of ProductTypes tag != 1");
        }
        nl2 = ((Element)nl.item(0)).getElementsByTagName("ProductType");
        for(int i=0; i<nl2.getLength(); ++i) {
        	Node node = nl2.item(i);
        	String name = node.getAttributes().getNamedItem("name").getNodeValue();
        	NodeList nl3 = ((Element)node).getElementsByTagName("Parameter");
        	Map<String, Double> map = new HashMap<String, Double>();
        	for(int j=0; j<nl3.getLength(); ++j) {
        		Node node2 = nl3.item(j);
        		String name2 = node2.getAttributes().getNamedItem("name").getNodeValue();
        		String value = node2.getAttributes().getNamedItem("value").getNodeValue();
        		map.put(name2.toLowerCase(), Double.parseDouble(value));
        	}
        	
        	mapProductTypes.put(name.toLowerCase(), map);
        }
 
        listFundNumber = new ArrayList<Integer>();
        listFundFee = new ArrayList<Double>();
        nl = doc.getElementsByTagName("Funds");
        nl2 = ((Element)nl.item(0)).getElementsByTagName("Fund");
        for(int i=0; i<nl2.getLength(); ++i) {
        	Node node = nl2.item(i);
        	String number = node.getAttributes().getNamedItem("number").getNodeValue();
        	String fee = node.getAttributes().getNamedItem("fee").getNodeValue();        	
        	
        	listFundNumber.add(Integer.parseInt(number));
        	listFundFee.add(Double.parseDouble(fee) / 10000.0);
        }
	}
	
	protected double getDoubleParam(String name) throws Exception {
		String key = name.toLowerCase();
		if(mapParameters.containsKey(key)) {
			return Double.parseDouble(mapParameters.get(key));
		} else {
			throw new Exception("Cannot find parameter: " + name);
		}
	}
	
	protected int getIntParam(String name) throws Exception {
		String key = name.toLowerCase();
		if(mapParameters.containsKey(key)) {
			return Integer.parseInt(mapParameters.get(key));
		} else {
			throw new Exception("Cannot find parameter: " + name);
		}
	}
	
	protected String getParam(String name) throws Exception {
		String key = name.toLowerCase();
		if(mapParameters.containsKey(key)) {
			return mapParameters.get(key);
		} else {
			throw new Exception("Cannot find parameter: " + name);
		}
	}

	protected SDate getDateParam(String name) throws Exception {
		String key = name.toLowerCase();
		if(mapParameters.containsKey(key)) {
			SimpleDateFormat df = new SimpleDateFormat("dd/M/yyyy");
			Calendar cal = Calendar.getInstance();
			cal.setTime(df.parse(mapParameters.get(key)));
			return new SDate(cal.get(Calendar.YEAR), cal.get(Calendar.MONTH) + 1, cal.get(Calendar.DAY_OF_MONTH));
		} else {
			throw new Exception("Cannot find parameter: " + name);
		}
	}
	
	protected double getProductTypeParam(String gtype, String pname) throws Exception {
		String key1 = gtype.toLowerCase();
		String key2 = pname.toLowerCase();
		if(mapProductTypes.containsKey(key1)) {
			Map<String, Double> map = mapProductTypes.get(key1);
			if(map.containsKey(key2)) {
				return map.get(key2);
			} else {
				throw new Exception("Cannot find ProductType value: " + gtype + ":" + pname);
			}
		} else {
			throw new Exception("Cannot find ProductType: " + gtype);
		}
	}
}
